/*--------------------------------*-C-*---------------------------------*
 * File:	strings.c
 *----------------------------------------------------------------------*
 *
 * All portions of code are copyright by their respective author/s.
 * Copyright (c) 1997-2001   Geoff Wing <gcw@pobox.com>
 * Copyright (c) 2004        Jingmin Zhou <jimmyzhou@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------*/
/*
** $Id: strings.c,v 1.16 2004/11/11 19:58:55 cvs Exp $
*/

#include "../config.h"
#include "rxvt.h"


#ifdef DEBUG_VERBOSE
#define DEBUG_LEVEL 1
#else 
#define DEBUG_LEVEL 0
#endif

#if DEBUG_LEVEL
#define DBG_MSG(d,x) if(d <= DEBUG_LEVEL) fprintf x
#else
#define DBG_MSG(d,x)
#endif


/*----------------------------------------------------------------------*/
/*
 * a replacement for strcasecmp() to avoid linking an entire library.
 * Mark Olesen added this in 2.15 but for which OS & library? - Geoff Wing
 */
/* EXTPROTO */
int
ma_strcasecmp(const char *s1, const char *s2)
{
    for ( ; tolower(*s1) == tolower(*s2); s1++, s2++)
	if (!*s1)
	    return 0;
    return (int)(tolower(*s1) - tolower(*s2));
}

/* EXTPROTO */
int
ma_strncasecmp(const char *s1, const char *s2, size_t n)
{
    for ( ; n-- && (tolower(*s1) == tolower(*s2)); s1++, s2++)
	if (!*s1)
	    return 0;
    if (n == 0)
	return 0;
    return (int)(tolower(*s1) - tolower(*s2));
}

/* EXTPROTO */
char*
ma_strcpy(char *d, const char *s)
{
    char          *r = d;

    for ( ; (*r++ = *s++) != '\0'; )
		;
    return d;
}

/* EXTPROTO */
char*
ma_strncpy(char *d, const char* s, size_t len)
{
    char          *r = d;

    if (len)
		for ( ; len; len--)
		    if ((*r++ = *s++) == '\0') {
				for ( ; --len; )
				    *r++ = '\0';
				break;
		    }
    return d;
}

/* EXTPROTO */
int
ma_strcmp(const char *s1, const char *s2)
{
    for ( ; (*s1 == *s2++); )
		if (*s1++ == '\0')
		    return 0;
    return (int) ((unsigned char) *s1 - (unsigned char) *--s2);
}

/* EXTPROTO */
int
ma_strncmp(const char *s1, const char *s2, size_t len)
{
    if (len) {
		for ( ; len-- && (*s1++ == *s2++); )
			;
		if (++len)
		    return (int) ((unsigned char) *--s1 - (unsigned char) *--s2);
    }
    return 0;
}

/* EXTPROTO */
char*
ma_strcat(char *s1, const char *s2)
{
    char           *r = s1;

    if (*r != '\0')
		for ( ; *++r != '\0'; )
			;
    for ( ; (*r++ = *s2++) != '\0'; )
		;

    return s1;
}

/* EXTPROTO */
char*
ma_strncat(char *s1, const char *s2, size_t len)
{
    char           *r = s1;

    if (*r != '\0')
		for ( ; *++r != '\0'; )
			;
    for ( ; len-- && ((*r++ = *s2++) != '\0'); )
		;
    *r = '\0';

    return s1;
}

/* EXTPROTO */
size_t
ma_strlen(const char *s)
{
    size_t         len = 0;

    for ( ; *s++ != '\0'; len++)
		;
    return len;
}


/* EXTPROTO */
char*
ma_strdup(const char *s)
{
	int			len = STRLEN(s) + 1;
    char*		c;

	assert (len > 0);	/* possible integer overflow? */
    if ((c = rxvt_malloc(len)) != NULL)
		MEMCPY(c, s, len);
    return c;
}


/* EXTPROTO */
$untainted char*
ma_strndup($tainted const char* s, size_t sz)
{
    char*		c;
	int			len = sz + 1;

	assert (len > 0);	/* possible integer overflow? */
    if ((c = rxvt_malloc(len)) != NULL)
		STRNCPY(c, s, sz);
	c[sz] = '\0';
    return c;
}


/* EXTPROTO */
char*
ma_index(const char* s, int c)
{
    return STRCHR(s, c);
}

/* EXTPROTO */
char*
ma_strchr(const char* s, int c)
{
    char          *p = NULL;

    for (;;) {
		if (*s == (char)c) {
		    p = (char *)s;
		    break;
		}
		if (*s++ == '\0')
		    break;
    }
    return p;

}

/* EXTPROTO */
char*
ma_rindex(const char* s, int c)
{
    return STRRCHR(s, c);
}

/* EXTPROTO */
char*
ma_strrchr(const char* s, int c)
{
    char          *p = NULL;

    for (;;) {
		if (*s == (char)c)
		    p = (char *)s;
		if (*s++ == '\0')
		    break;
    }
    return p;
}

/* EXTPROTO */
void*
ma_memcpy(void *s1, const void *s2, size_t len)
{
    /* has extra stack and time but less code space */
    return MEMMOVE(s1, s2, len);
}

/*--------------------------------------------------------------------------*
 * Possibly faster memmove() by Geoff Wing <mason@primenet.com.au>
 *--------------------------------------------------------------------------*/
/* EXTPROTO */
void*
ma_memmove(void *d, const void *s, size_t len)
{
    u_intp_t        i;
	u_intp_t*		pdst;
	u_intp_t*		psrc;
    unsigned char*	dst = (unsigned char *)d;
    unsigned char*	src = (unsigned char *)s;

    if (len && d != s) {
		if ((u_intp_t)d < (u_intp_t)s) {
			/* forwards */
		    i = (-(u_intp_t)dst) & (SIZEOF_INT_P - 1);
		    if (len >= 16 &&
				i == ((-(u_intp_t)src) & (SIZEOF_INT_P - 1))) {
			    /* speed up since src & dst are offset correctly */
				len -= (size_t)i;
				for ( ; i--; )
				    *dst++ = *src++;
				/* assign the src/dst to psrc/pdst */
				pdst = (u_intp_t*) dst;
				psrc = (u_intp_t*) src;
				for (i = (u_intp_t)(len / SIZEOF_INT_P); i--; )
					*pdst++ = *psrc++;
				len &= (SIZEOF_INT_P - 1);
				/* assign back the src/dst */
				dst = (unsigned char*) pdst;
				src = (unsigned char*) psrc;
			}
			/* the left bytes */
		    for ( ; len--; )
				*dst++ = *src++;
		}
		else {
			/* backwards */
		    dst += len;
		    src += len;
		    i = ((u_intp_t)dst) & (SIZEOF_INT_P - 1);
		    if (len >= 16 &&
				i == (((u_intp_t)src) & (SIZEOF_INT_P - 1))) {
			    /* speed up since src & dst are offset correctly */
				len -= (size_t)i;
				for ( ; i--; )
				    *--dst = *--src;
				/* assign the src/dst to psrc/pdst */
				pdst = (u_intp_t*) dst;
				psrc = (u_intp_t*) src;
				for (i = (u_intp_t)(len / SIZEOF_INT_P); i--; )
				    *--pdst = *--psrc;
				len &= (SIZEOF_INT_P - 1);
				/* assign back the src/dst */
				dst = (unsigned char*) pdst;
				src = (unsigned char*) psrc;
			}
			/* the left bytes */
		    for ( ; len--; )
				*--dst = *--src;
		}
    }
    return d;
}

/*--------------------------------------------------------------------------*
 * Possibly faster memset() by Geoff Wing <mason@primenet.com.au>
 * presumptions:
 *   1) intp_t write the best
 *   2) SIZEOF_INT_P == power of 2
 *--------------------------------------------------------------------------*/

/* EXTPROTO */
void
ma_bzero(void* buf, size_t len)
{
    MEMSET(buf, 0, len);
}

/* EXTPROTO */
void*
ma_memset(void *p, int c1, size_t len)
{
    u_intp_t        i, val;
	u_intp_t*		pdst;
    unsigned char   c = (unsigned char) c1;
    unsigned char  *lp = (unsigned char *) p;

    if (len) {
		if (len >= 16) {
			/*
			** < 16 probably not worth all the calculations
			** write out preceding characters so we align on an
			** integer boundary
			*/
		    if ((i = ((-(u_intp_t)p) & (SIZEOF_INT_P - 1)))) {
				len -= (size_t)i;
				for (; i--;)
				    *lp++ = c;
		    }

			/* do the fast writing */
		    val = (c << 8) + c;
#if SIZEOF_INT_P >= 4
		    val |= (val << 16);
#endif
#if SIZEOF_INT_P >= 8
		    val |= (val << 32);
#endif
#if SIZEOF_INT_P == 16
		    val |= (val << 64);
#endif
			/* assign the lp to pdst */
			pdst = (u_intp_t*) lp;
		    for (i = (u_intp_t)(len / SIZEOF_INT_P); i--;)
				*pdst++ = val;
		    len &= (SIZEOF_INT_P - 1);
			/* assign back the pdst to lp */
			lp = (unsigned char*) pdst;
		}
		/* write trailing characters */
		for (; len--;)
		    *lp++ = c;
    }
    return p;
}
/*----------------------- end-of-file (C source) -----------------------*/
